import logging
from collections import Counter

from pastml import get_personalized_feature_name, METHOD, STATES, CHARACTER, NUM_SCENARIOS, NUM_UNRESOLVED_NODES, \
    NUM_NODES, NUM_TIPS, NUM_STATES_PER_NODE, PERC_UNRESOLVED

STEPS = 'steps'

DOWNPASS = 'DOWNPASS'
ACCTRAN = 'ACCTRAN'
DELTRAN = 'DELTRAN'
MP = 'MP'

MP_METHODS = {DOWNPASS, ACCTRAN, DELTRAN}
META_MP_METHODS = {MP}

BU_PARS_STATES = 'BOTTOM_UP_PARSIMONY'
TD_PARS_STATES = 'TOP_DOWN_PARSIMONY'
PARS_STATES = 'PARSIMONY'
PARS_STATE2NUM = 'PARSIMONY_STEPS'


def is_meta_mp(method):
    """
    Checks if the method is a meta max parsimony method, combining several methods, i.e. MP.

    :param method: prediction method
    :type method: str
    :return: bool
    """
    return method in META_MP_METHODS


def get_default_mp_method():
    return DOWNPASS


def is_parsimonious(method):
    """
    Checks if the method is max likelihood, i.e. is either joint or one of the marginal ones
    (marginal itself, or MAP, or MPPA).

    :param method: the ancestral state prediction method used by PastML.
    :type method: str
    :return: whether the method is parsimonious
    :rtype: bool
    """
    return method in MP_METHODS | {MP}


def initialise_parsimonious_states(tree, feature, states):
    """
    Initializes the bottom-up state arrays for tips based on their states given by the feature.

    :param tree: the tree of interest
    :type tree: ete3.Tree
    :param feature: feature in which the tip states are stored (the value could be None for a missing state)
    :type feature: str
    :param states: possible character states
    :type states: np.array(str)
    :return: adds the get_personalised_feature_name(feature, BU_PARS) feature to tree tips.
    """
    ps_feature_down = get_personalized_feature_name(feature, BU_PARS_STATES)
    ps_feature = get_personalized_feature_name(feature, PARS_STATES)
    all_states = set(states)

    for node in tree.traverse():
        state = getattr(node, feature, set())
        if not state:
            node.add_feature(ps_feature_down, all_states)
        else:
            node.add_feature(ps_feature_down, state)
        node.add_feature(ps_feature, getattr(node, ps_feature_down))


def get_most_common_states(state_iterable):
    """
    Gets the set of most common states among the state sets contained in the iterable argument

    :param state_iterable: iterable of state sets
    :type state_iterable: iterable
    :return: set of most common states
    :rtype: set(str)
    """
    state_counter = Counter()
    for states in state_iterable:
        state_counter.update(states)
    max_count = state_counter.most_common(1)[0][1]
    return {state for (state, count) in state_counter.items() if count == max_count}


def uppass(tree, feature):
    """
    UPPASS traverses the tree starting from the tips and going up till the root,
    and assigns to each parent node a state based on the states of its child nodes.

    if N is a tip:
    S(N) <- state of N
    else:
      L, R <- left and right children of N
      UPPASS(L)
      UPPASS(R)
      if S(L) intersects with S(R):
         S(N) <- intersection(S(L), S(R))
      else:
         S(N) <- union(S(L), S(R))

    :param tree: the tree of interest
    :type tree: ete3.Tree
    :param feature: character for which the parsimonious states are reconstructed
    :type feature: str
    :return: adds get_personalized_feature_name(feature, BU_PARS_STATES) feature to the tree nodes
    """

    ps_feature = get_personalized_feature_name(feature, BU_PARS_STATES)

    for node in tree.traverse('postorder'):
        if not node.is_leaf():
            children_states = get_most_common_states(getattr(child, ps_feature) for child in node.children)
            node_states = getattr(node, ps_feature)
            state_intersection = node_states & children_states
            node.add_feature(ps_feature, state_intersection if state_intersection else node_states)


def acctran(tree, character, feature=PARS_STATES):
    """
    ACCTRAN (accelerated transformation) (Farris, 1970) aims at reducing the number of ambiguities
    in the parsimonious result. ACCTRAN forces the state changes to be performed as close to the root as possible,
    and therefore prioritises the reverse mutations.

    if N is not a tip:
        L, R <- left and right children of N
        if intersection(S(N), S(L)) is not empty:
            S(L) <- intersection(S(N), S(L))
        if intersection(S(N), S(R)) is not empty:
            S(R) <- intersection(S(N), S(R))
        ACCTRAN(L)
        ACCTRAN(R)

    :param tree: the tree of interest
    :type tree: ete3.Tree
    :param character: character for which the parsimonious states are reconstructed
    :type character: str
    :param feature: feature where the reconstructed states are to be saved
    :type feature: str
    :return: adds get_personalized_feature_name(feature, PARS_STATES) feature to the tree nodes
    """

    ps_feature_down = get_personalized_feature_name(character, BU_PARS_STATES)

    for node in tree.traverse('preorder'):
        if node.is_root():
            node.add_feature(feature, getattr(node, ps_feature_down))
        node_states = getattr(node, feature)
        for child in node.children:
            child_states = getattr(child, ps_feature_down)
            state_intersection = node_states & child_states
            child.add_feature(feature, state_intersection if state_intersection else child_states)


def downpass(tree, feature, states):
    """
    DOWNPASS traverses the tree starting from the root and going down till the tips,
    and for each node combines the state information from its supertree and its subtree (calculated at UPPASS).
    As the root state was already the most parsimonious after the UPPASS,
    we skip it and start directly with the root children.

    if N is not a tip:
        L, R <- left and right children of N
        if N is root:
            UP_S(N) <- union of all states
        else:
            P <- parent of N
            B <- brother of N
            UP_S(N) <- most_common_states(UP_S(P), S(B))
        S(N) <- most_common_states(UP_S(N), S(L), S(R))
        DOWNPASS(L)
        DOWNPASS(R)

    :param tree: the tree of interest
    :type tree: ete3.Tree
    :param feature: character for which the parsimonious states are reconstructed
    :type feature: str
    :param states: possible character states
    :type states: np.array(str)
    :return: adds get_personalized_feature_name(feature, PARS_STATES) feature to the tree nodes
    """

    ps_feature_down = get_personalized_feature_name(feature, BU_PARS_STATES)
    ps_feature_up = get_personalized_feature_name(feature, TD_PARS_STATES)
    ps_feature = get_personalized_feature_name(feature, PARS_STATES)

    for node in tree.traverse('preorder'):
        if node.is_root():
            node.add_feature(ps_feature_up, set(states))
        else:
            node.add_feature(ps_feature_up,
                             get_most_common_states([getattr(node.up, ps_feature_up)]
                                                    + [getattr(sibling, ps_feature_down) for sibling in node.up.children
                                                       if sibling != node]))
        down_up_states = get_most_common_states([getattr(node, ps_feature_up)]
                                                + [getattr(child, ps_feature_down) for child in node.children]) \
            if not node.is_leaf() else getattr(node, ps_feature_up)
        preset_states = getattr(node, ps_feature)
        state_intersection = down_up_states & preset_states
        node.add_feature(ps_feature, state_intersection if state_intersection else preset_states)

    for node in tree.traverse():
        node.del_feature(ps_feature_down)
        node.del_feature(ps_feature_up)


def deltran(tree, feature):
    """
    DELTRAN (delayed transformation) (Swofford & Maddison, 1987) aims at reducing the number of ambiguities
    in the parsimonious result. DELTRAN makes the changes as close as possible to the leaves,
    hence prioritizing parallel mutations. DELTRAN is performed after DOWNPASS.

    if N is not a root:
        P <- parent(N)
        if intersection(S(N), S(P)) is not empty:
            S(N) <- intersection(S(N), S(P))
    if N is not a tip:
        L, R <- left and right children of N
        DELTRAN(L)
        DELTRAN(R)

    :param tree: the tree of interest
    :type tree: ete3.Tree
    :param feature: character for which the parsimonious states are reconstructed
    :type feature: str
    :return: modifies get_personalized_feature_name(feature, PARS_STATES) feature of the tree nodes
    """
    ps_feature = get_personalized_feature_name(feature, PARS_STATES)

    for node in tree.traverse('preorder'):
        if not node.is_root():
            node_states = getattr(node, ps_feature)
            parent_states = getattr(node.up, ps_feature)
            state_intersection = node_states & parent_states
            if state_intersection:
                node.add_feature(ps_feature, state_intersection)


def parsimonious_acr(forest, character, prediction_method, states, num_nodes, num_tips):
    """
    Calculates parsimonious states on the trees and stores them in the corresponding feature.

    :param forest: trees of interest
    :type forest: list(ete3.Tree)
    :param character: character for which the parsimonious states are reconstructed
    :type character: str
    :param prediction_method: ACCTRAN (accelerated transformation), DELTRAN (delayed transformation), DOWNPASS or MP
    :type prediction_method: str
    :param states: possible character states
    :type states: np.array(str)
    :param num_nodes: total number of nodes in the forest
    :type num_nodes: int
    :param num_tips: total number of tips in the forest
    :type num_tips: int
    :return: mapping between reconstruction parameters and values
    :rtype: dict
    """
    for tree in forest:
        initialise_parsimonious_states(tree, character, states)
        uppass(tree, character)

    results = []
    result = {STATES: states, NUM_NODES: num_nodes, NUM_TIPS: num_tips}

    logger = logging.getLogger('pastml')

    def process_result(method, feature):
        out_feature = get_personalized_feature_name(character, method) if prediction_method != method else character
        res = result.copy()
        res[NUM_SCENARIOS], res[NUM_UNRESOLVED_NODES], res[NUM_STATES_PER_NODE] = 1, 0, 0
        for tree in forest:
            ns, nun, nspn = choose_parsimonious_states(tree, feature, out_feature)
            res[NUM_SCENARIOS] *= ns
            res[NUM_UNRESOLVED_NODES] += nun
            res[NUM_STATES_PER_NODE] += nspn
        res[NUM_STATES_PER_NODE] /= num_nodes
        res[PERC_UNRESOLVED] = res[NUM_UNRESOLVED_NODES] * 100 / num_nodes
        logger.debug('{} node{} unresolved ({:.2f}%) for {} by {}, '
                     'i.e. {:.4f} state{} per node in average.'
                     .format(res[NUM_UNRESOLVED_NODES], 's are' if res[NUM_UNRESOLVED_NODES] != 1 else ' is',
                             res[PERC_UNRESOLVED], character, method,
                             res[NUM_STATES_PER_NODE], 's' if res[NUM_STATES_PER_NODE] > 1 else ''))
        res[CHARACTER] = out_feature
        res[METHOD] = method
        results.append(res)

    if prediction_method in {ACCTRAN, MP}:
        feature = get_personalized_feature_name(character, PARS_STATES)
        if prediction_method == MP:
            feature = get_personalized_feature_name(feature, ACCTRAN)
        result[STEPS] = 0
        for tree in forest:
            acctran(tree, character, feature)
            result[STEPS] += get_num_parsimonious_steps(tree, feature)
        process_result(ACCTRAN, feature)

        bu_feature = get_personalized_feature_name(character, BU_PARS_STATES)
        for tree in forest:
            for node in tree.traverse():
                if prediction_method == ACCTRAN:
                    node.del_feature(bu_feature)
                node.del_feature(feature)

    if prediction_method != ACCTRAN:
        feature = get_personalized_feature_name(character, PARS_STATES)
        result[STEPS] = 0
        for tree in forest:
            downpass(tree, character, states)
            if prediction_method in {DOWNPASS, MP}:
                result[STEPS] += get_num_parsimonious_steps(tree, feature)
        if prediction_method in {DOWNPASS, MP}:
            process_result(DOWNPASS, feature)
        result[STEPS] = 0
        if prediction_method in {DELTRAN, MP}:
            for tree in forest:
                deltran(tree, character)
                result[STEPS] += get_num_parsimonious_steps(tree, feature)
            process_result(DELTRAN, feature)
        for tree in forest:
            for node in tree.traverse():
                node.del_feature(feature)

    logger.debug("Parsimonious reconstruction for {} requires {} state changes."
                 .format(character, result[STEPS]))
    return results


def choose_parsimonious_states(tree, ps_feature, out_feature):
    """
    Calculates the statistics for parsimony, and copies the parsimonious states to out_feature.

    :param ps_feature: feature containing each node parsimonious states
    :type ps_feature: str
    :param out_feature: feature where the node parsimonious states are to be copied
    :type out_feature: str
    :param tree: the tree of interest
    :type tree: ete3.Tree
    :return: number of ancestral scenarios selected, number of unresolved nodes, and total numbest of selected states
    :rtype: tuple(int, int, int)
    """
    num_scenarios = 1
    unresolved_nodes = 0
    num_states = 0
    for node in tree.traverse():
        states = getattr(node, ps_feature)
        node.add_feature(out_feature, states)
        n = len(states)
        num_scenarios *= n
        unresolved_nodes += 1 if n > 1 else 0
        num_states += n
    return num_scenarios, unresolved_nodes, num_states


def get_num_parsimonious_steps(tree, feature):
    ps_feature_num = get_personalized_feature_name(feature, PARS_STATE2NUM)

    for node in tree.traverse('postorder'):
        if node.is_leaf():
            node.add_feature(ps_feature_num, {state: 0 for state in getattr(node, feature)})
        else:
            state2num = {}
            for state in getattr(node, feature):
                num = 0
                for child in node.children:
                    child_state2num = getattr(child, ps_feature_num)
                    num += min(((0 if state == child_state else 1) + child_num)
                               for (child_state, child_num) in child_state2num.items())
                state2num[state] = num
            node.add_feature(ps_feature_num, state2num)
            for child in node.children:
                child.del_feature(ps_feature_num)
    state2num = getattr(tree, ps_feature_num)
    tree.del_feature(ps_feature_num)
    return min(state2num.values())



import numpy as np


def save_matrix(states, matrix, outfile):
    """
    Saves a matrix into a file.

    :param states: names of the columns/rows, will be saved in a commented (starting with #) header line before the matrix
    :param matrix: np.array matrix to be saved
    :param outfile: path to the file where the matrix should be saved
    :return: void
    """
    np.savetxt(outfile, matrix, delimiter=' ', fmt='%.18e', header=' '.join(states))


def get_diagonalisation(frequencies, rate_matrix=None):
    """
    Normalises and diagonalises the rate matrix.

    :param frequencies: character state frequencies.
    :type frequencies: numpy.array
    :param rate_matrix: (optional) rate matrix (by default an all-equal-rate matrix is used)
    :type rate_matrix: numpy.ndarray
    :return: matrix diagonalisation (d, A, A^{-1})
        such that A.dot(np.diag(d))).dot(A^{-1}) = 1/mu Q (normalised generator)
    :rtype: tuple
    """
    Q = get_normalised_generator(frequencies, rate_matrix)
    d, A = np.linalg.eig(Q)
    return d, A, np.linalg.inv(A)


def get_normalised_generator(frequencies, rate_matrix=None):
    """
    Calculates the normalised generator from the rate matrix and character state frequencies.

    :param frequencies: character state frequencies.
    :type frequencies: numpy.array
    :param rate_matrix: (optional) rate matrix (by default an all-equal-rate matrix is used)
    :type rate_matrix: numpy.ndarray
    :return: normalised generator 1/mu Q
    :rtype: numpy.ndarray
    """
    if rate_matrix is None:
        n = len(frequencies)
        rate_matrix = np.ones(shape=(n, n), dtype=np.float64) - np.eye(n)
    generator = rate_matrix * frequencies
    generator -= np.diag(generator.sum(axis=1))
    mu = -generator.diagonal().dot(frequencies)
    generator /= mu
    return generator


def get_pij_matrix(t, diag, A, A_inv):
    """
    Calculates the probability matrix of substitutions i->j over time t,
    given the normalised generator diagonalisation.


    :param t: time
    :type t: float
    :return: probability matrix
    :rtype: numpy.ndarray
    """
    return A.dot(np.diag(np.exp(diag * t))).dot(A_inv)




import logging

import numpy as np

from pastml.models import Model, ModelWithFrequencies

HKY = 'HKY'
HKY_STATES = np.array(['A', 'C', 'G', 'T'])
A, C, G, T = 0, 1, 2, 3
KAPPA = 'kappa'


class HKYModel(ModelWithFrequencies):

    def __init__(self, forest_stats, sf=None, frequencies=None, kappa=4, tau=0,
                 frequency_smoothing=False, optimise_tau=False, parameter_file=None, reoptimise=False, **kwargs):
        self._kappa = None
        self._optimise_kappa = True
        kwargs['states'] = HKY_STATES
        ModelWithFrequencies.__init__(self, forest_stats=forest_stats,
                                      sf=sf, tau=tau, optimise_tau=optimise_tau, frequencies=frequencies,
                                      frequency_smoothing=frequency_smoothing, reoptimise=reoptimise,
                                      parameter_file=parameter_file, **kwargs)
        if self._kappa is None:
            self._kappa = kappa
        self.name = HKY

    @property
    def kappa(self):
        return self._kappa

    @kappa.setter
    def kappa(self, kappa):
        if self._optimise_kappa:
            self._kappa = kappa
        else:
            raise NotImplementedError('The kappa value is preset and cannot be changed.')

    @Model.states.setter
    def states(self, states):
        raise NotImplementedError("The HKY model is only implemented for nucleotides: "
                                  "the states are A, C, G, T and cannot be reset")

    def get_Pij_t(self, t, *args, **kwargs):
        """
        Calculates the probability matrix of substitutions i->j over time t,
        with HKY model [Hasegawa, Kishino and Yano 1985], given state frequencies and kappa.
        Is only implemented for 4 nucleotide states.

        :param t: time
        :type t: float
        :return: probability matrix
        :rtype: numpy.ndarray
        """
        t = self.transform_t(t)

        pi_a, pi_c, pi_g, pi_t = self.frequencies
        pi_ag = pi_a + pi_g
        pi_ct = pi_c + pi_t
        beta = .5 / (pi_ag * pi_ct + self.kappa * (pi_a * pi_g + pi_c * pi_t))

        exp_min_beta_t = np.exp(-beta * t)
        exp_ct = np.exp(-beta * t * (1. + pi_ct * (self.kappa - 1.))) / pi_ct
        exp_ag = np.exp(-beta * t * (1. + pi_ag * (self.kappa - 1.))) / pi_ag
        ct_sum = (pi_ct + pi_ag * exp_min_beta_t) / pi_ct
        ag_sum = (pi_ag + pi_ct * exp_min_beta_t) / pi_ag
        p = np.ones((4, 4), dtype=np.float64) * (1 - exp_min_beta_t)
        p *= self.frequencies

        p[T, T] = pi_t * ct_sum + pi_c * exp_ct
        p[T, C] = pi_c * ct_sum - pi_c * exp_ct

        p[C, T] = pi_t * ct_sum - pi_t * exp_ct
        p[C, C] = pi_c * ct_sum + pi_t * exp_ct

        p[A, A] = pi_a * ag_sum + pi_g * exp_ag
        p[A, G] = pi_g * ag_sum - pi_g * exp_ag

        p[G, A] = pi_a * ag_sum - pi_a * exp_ag
        p[G, G] = pi_g * ag_sum + pi_a * exp_ag

        return p

    def get_num_params(self):
        """
        Returns the number of optimized parameters for this model.

        :return: the number of optimized parameters
        """
        return ModelWithFrequencies.get_num_params(self) + (1 if self._optimise_kappa else 0)

    def set_params_from_optimised(self, ps, **kwargs):
        """
        Update this model parameter values from a vector representing parameters
        for the likelihood optimization algorithm.

        :param ps: np.array containing parameters of the likelihood optimization algorithm
        :param kwargs: dict of eventual other arguments
        :return: void, update this model
        """
        if self.extra_params_fixed():
            Model.set_params_from_optimised(self, ps, **kwargs)
        else:
            ModelWithFrequencies.set_params_from_optimised(self, ps, **kwargs)
            n_params = ModelWithFrequencies.get_num_params(self)
            if self._optimise_kappa:
                self.kappa = ps[n_params]

    def get_optimised_parameters(self):
        """
        Converts this model parameters to a vector representing parameters
        for the likelihood optimization algorithm.

        :return: np.array containing parameters of the likelihood optimization algorithm
        """
        if self.extra_params_fixed():
            return Model.get_optimised_parameters(self)
        return np.hstack((ModelWithFrequencies.get_optimised_parameters(self),
                          [self.kappa] if self._optimise_kappa else []))

    def get_bounds(self):
        """
        Get bounds for parameters for likelihood optimization algorithm.

        :return: np.array containing lower and upper (potentially infinite) bounds for each parameter
        """
        if self.extra_params_fixed():
            return Model.get_bounds(self)
        return np.array((*ModelWithFrequencies.get_bounds(self),
                        *([np.array([1e-6, 20.])] if self._optimise_kappa else [])))

    def parse_parameters(self, params, reoptimise=False):
        """
        Update this model's values from the input parameters.
        For the HKY model, apart from the basic parameters (scaling factor and smoothing factor) and the frequencies
        (see pastml.models.ModelWithFrequencies),
        the input might contain an optional kappa value:
        the key for kappa value is the pastml.models.HKYModel.KAPPA.

        :param params: dict {key->value}
        :param reoptimise: whether these model parameters should be treated as starting values (True)
            or as fixed values (False)
        :return: dict with parameter values (same as input)
        """
        params = ModelWithFrequencies.parse_parameters(self, params, reoptimise)
        logger = logging.getLogger('pastml')
        if KAPPA in params:
            self._kappa = params[KAPPA]
            try:
                self._kappa = np.float64(self._kappa)
                if self._kappa <= 0:
                    logger.error(
                        'Kappa cannot be negative, ignoring the value given in paramaters ({}).'.format(self._kappa))
                    self._kappa = None
                else:
                    self._optimise_kappa = reoptimise
            except:
                logger.error('Kappa ({}) given in parameters is not float, ignoring it.'.format(self._kappa))
                self._kappa = None
        return params

    def _print_parameters(self):
        """
        Constructs a string representing parameter values (to be used to logging).

        :return: str representing parameter values
        """
        return '{}' \
               '\tkappa\t{:.6f}\t{}\n'.format(ModelWithFrequencies.get_optimised_parameters(self),
                                          self.kappa, '(optimised)' if self._optimise_kappa else '(fixed)')

    def freeze(self):
        """
        Prohibit parameter optimization by setting all optimization flags to False.

        :return: void
        """
        ModelWithFrequencies.freeze(self)
        self._optimise_kappa = False

    def save_parameters(self, filehandle):
        """
        Writes this model parameter values to the parameter file (in the same format as the input parameter file).

        :param filehandle: filehandle for the file where the parameter values should be written.
        :return: void
        """
        ModelWithFrequencies.save_parameters(self, filehandle)
        filehandle.write('{}\t{:g}\n'.format(KAPPA, self.kappa))



import logging
import os
from pathlib import Path

import pandas as pd

from pastml.ml import is_marginal, MARGINAL_PROBABILITIES, MODEL
from pastml import METHOD, CHARACTER, STATES
from pastml.visualisation.colour_generator import get_enough_colours, parse_colours
from itolapi import Itol

from pastml.tree import DATE

STYLE_FILE_HEADER_TEMPLATE = """DATASET_STYLE

SEPARATOR TAB
DATASET_LABEL	{column} ({method})
COLOR	#ffffff

LEGEND_COLORS	{colours}
LEGEND_LABELS	{states}
LEGEND_SHAPES	{shapes}
LEGEND_TITLE	{column} ({method})

DATA
#NODE_ID TYPE   NODE  COLOR SIZE_FACTOR LABEL_OR_STYLE
"""

COLORSTRIP_FILE_HEADER_TEMPLATE = """DATASET_COLORSTRIP

SEPARATOR TAB
BORDER_WIDTH	0.5
MARGIN	5
STRIP_WIDTH	25
COLOR	#ffffff
DATASET_LABEL	{column}

LEGEND_COLORS	{colours}
LEGEND_LABELS	{states}
LEGEND_SHAPES	{shapes}
LEGEND_TITLE	{column}

DATA
#NODE_ID COLOR LABEL_OR_STYLE
"""

POPUP_FILE_HEADER = """POPUP_INFO

SEPARATOR TAB

DATA
#NODE_ID POPUP_TITLE POPUP_CONTENT
"""

POPUP_CONTENT_TEMPLATE = "<b>{key}: </b>" \
                         "<div style='overflow:auto;max-width:50vw;'>" \
                         "<span style='white-space:nowrap;'>{value}</span></div>"

DEFAULT_ITOL_PROJECT = 'Sample project'


def generate_itol_annotations(column2states, work_dir, acrs, state_df, date_col,
                              tree_path, itol_id=None, itol_project=None, itol_tree_name=None,
                              column2colours=None):
    annotation_files = []
    popup_file = os.path.join(work_dir, 'iTOL_popup_info.txt')
    with open(popup_file, 'w+') as pf:
        pf.write(POPUP_FILE_HEADER)

    state_df['itol_type'] = 'branch'
    state_df['itol_node'] = 'node'
    state_df['itol_style'] = 'normal'
    state_df['itol_size'] = 2

    for acr_result in acrs:
        column = acr_result[CHARACTER]
        states = acr_result[STATES]
        num_unique_values = len(states)
        if column2colours and column in column2colours:
            colours = parse_colours(column2colours[column], states)
        else:
            colours = get_enough_colours(num_unique_values)
        value2colour = dict(zip(states, colours))
        style_file = os.path.join(work_dir, 'iTOL_style-{}.txt'.format(column))
        with open(style_file, 'w+') as sf:
            sf.write(STYLE_FILE_HEADER_TEMPLATE
                     .format(column=column, colours='\t'.join(colours), states='\t'.join(states),
                             shapes='\t'.join(['1'] * len(states)),
                             method='{}{}'.format(acr_result[METHOD],
                                                  ('+{}'.format(acr_result[MODEL].name) if MODEL in acr_result else ''))))
        col_df = state_df[state_df[column].apply(len) == 1]
        col_df['itol_label'] = col_df[column].apply(lambda _: next(iter(_)))
        col_df['itol_colour'] = col_df['itol_label'].apply(lambda _: value2colour[_])
        col_df[['itol_type', 'itol_node', 'itol_colour', 'itol_size', 'itol_style']].to_csv(style_file, sep='\t',
                                                                                            header=False, mode='a')
        annotation_files.append(style_file)
        logging.getLogger('pastml').debug('Generated iTol style file for {}: {}.'.format(column, style_file))

        colorstrip_file = os.path.join(work_dir, 'iTOL_colorstrip-{}.txt'.format(column))
        with open(colorstrip_file, 'w+') as csf:
            csf.write(COLORSTRIP_FILE_HEADER_TEMPLATE.format(column=column, colours='\t'.join(colours),
                                                             states='\t'.join(states),
                                                             shapes='\t'.join(['1'] * len(states))))
        col_df[['itol_colour', 'itol_label']].to_csv(colorstrip_file, sep='\t', header=False, mode='a')
        annotation_files.append(colorstrip_file)
        logging.getLogger('pastml').debug('Generated iTol colorstrip file for {}: {}.'.format(column, colorstrip_file))

    state_df = state_df[list(column2states.keys()) + ['dist', DATE]]
    for c in column2states.keys():
        state_df[c] = state_df[c].apply(lambda _: ' or '.join(sorted(_)))
    state_df.columns = ['ACR {} predicted state'.format(c) for c in column2states.keys()] + ['Node dist', date_col]
    state_df['Node id'] = state_df.index

    for acr_result in acrs:
        if is_marginal(acr_result[METHOD]):
            df = acr_result[MARGINAL_PROBABILITIES]
            state_df.loc[df.index.map(str),
                         'ACR {character} marginal probabilities'.format(character=acr_result[CHARACTER])] = \
                df.apply(lambda vs: ', '.join(('{}: {:g}'.format(c, mp) for (c, mp) in zip(df.columns, vs))),
                         axis=1)
    cols = sorted(state_df.columns, reverse=True)
    state_df['popup_info'] = \
        state_df[cols].apply(lambda vs: '<br>'.join(((POPUP_CONTENT_TEMPLATE
                                                      if c.startswith('ACR ') else '<b>{key}: </b>: {value}')
                                                    .format(key=c, value=v) for (c, v) in zip(cols, vs)
                                                     if not pd.isna(v))),
                             axis=1)
    state_df['label'] = 'ACR results'
    state_df[['label', 'popup_info']].to_csv(popup_file, sep='\t', mode='a', header=False)
    annotation_files.append(popup_file)
    logging.getLogger('pastml').debug('Generated iTol pop-up file: {}.'.format(popup_file))

    if itol_id:
        if not itol_project:
            logging.getLogger('pastml').debug('Trying "{}" as iTOL project, the tree will be uploaded to. '
                                              'To upload to a different project, use itol_project argument.'
                                              .format(DEFAULT_ITOL_PROJECT))
        if not itol_tree_name:
            itol_tree_name = os.path.splitext(os.path.basename(tree_path))[0]
        tree_id, web_page = upload_to_itol(tree_path, annotation_files, tree_name=itol_tree_name,
                                           tree_description=None, project_name=itol_project, upload_id=itol_id)
        if web_page:
            with open(os.path.join(work_dir, 'iTOL_url.txt'), 'w+') as f:
                f.write(web_page)
        if tree_id:
            with open(os.path.join(work_dir, 'iTOL_tree_id.txt'), 'w+') as f:
                f.write(tree_id)
    else:
        logging.getLogger('pastml').debug('To upload your tree(s) to iTOL, please specify the itol_id argument.')


def upload_to_itol(tree_path, dataset_paths, tree_name=None, tree_description=None, project_name=None, upload_id=None):
    try:
        itol_uploader = Itol()
        itol_uploader.add_file(Path(tree_path))
        for annotation_file in dataset_paths:
            itol_uploader.add_file(Path(annotation_file))
        if tree_name:
            itol_uploader.params['treeName'] = tree_name
        if tree_description:
            itol_uploader.params['treeDescription'] = tree_description
        if upload_id:
            itol_uploader.params['uploadID'] = upload_id
            if project_name:
                itol_uploader.params['projectName'] = project_name
        if itol_uploader.upload():
            logging.getLogger('pastml').debug(
                'Successfully uploaded your tree ({}) to iTOL: {}.'.format(itol_uploader.comm.tree_id,
                                                                           itol_uploader.get_webpage()))
            return itol_uploader.comm.tree_id, itol_uploader.get_webpage()
        else:
            status = itol_uploader.comm.upload_output
    except Exception as e:
        status = e
    logging.getLogger('pastml').error(
        'Failed to upload your tree to iTOL because of "{}". Please check your internet connection and iTOL settings{}.'
            .format(status,
                    (', e.g. your iTOL batch upload id ({}){}'
                     .format(upload_id,
                             (' and whether the project {} exists'.format(project_name) if project_name else '')))
                    if upload_id else ''))
    return None, None



from collections import Counter

import numpy as np


def simulate_states(tree, model, character, n_repetitions=1_000):
    n_states = len(model.states)
    state_ids = np.array(range(n_states))

    get_pij = model.get_Pij_t

    for n in tree.traverse('levelorder'):
        if n.is_root():
            random_states = np.random.choice(state_ids, size=n_repetitions, p=model.frequencies)
        else:
            probs = get_pij(n.dist)
            probs = np.maximum(probs, 0)
            random_states = np.zeros(n_repetitions, dtype=int)
            parent_states = getattr(n.up, character)
            sorted_indices = np.argsort(parent_states)
            parent_state_nums = Counter(parent_states)
            offset = 0
            for i in state_ids:
                parent_nums_i = parent_state_nums[i]
                if parent_nums_i > 0:
                    random_states[sorted_indices[offset: offset + parent_nums_i]] = \
                        np.random.choice(state_ids, size=parent_nums_i, p=probs[i])
                    offset += parent_nums_i
        n.add_feature(character, random_states)

    return tree



import logging
from collections import defaultdict
from pastml.tree import IS_POLYTOMY, copy_forest

import numpy as np

VERTICAL = 'VERTICAL'
HORIZONTAL = 'HORIZONTAL'
TRIM = 'TRIM'

IS_TIP = 'is_tip'

REASONABLE_NUMBER_OF_TIPS = 15

CATEGORIES = 'categories'

NUM_TIPS_INSIDE = 'max_size'

TIPS_INSIDE = 'in_tips'
INTERNAL_NODES_INSIDE = 'in_ns'
TIPS_BELOW = 'all_tips'
ROOTS = 'roots'
ROOT_DATES = 'root_dates'

COMPRESSED_NODE = 'compressed_node'

METACHILD = 'metachild'

IN_FOCUS = 'in_focus'
AROUND_FOCUS = 'around_focus'
UP_FOCUS = 'up_focus'


def _tree2pajek_vertices_arcs(compressed_tree, nodes, edges, columns):
    n2id = {}

    def get_states(n, columns):
        res = []
        for column in columns:
            values = getattr(n, column, set())
            value = values if isinstance(values, str) else ' or '.join(sorted(values))
            res.append('{}:{}'.format(column, value))
        return res

    for id, n in enumerate(compressed_tree.traverse('preorder'), start=len(nodes) + 1):
        n2id[n] = id
        if not n.is_root():
            edges.append('{} {} {}'.format(n2id[n.up], id, len(getattr(n, ROOTS))))
        nodes.append('{} "{}" "{}" {}'.format(id, n.name,
                                              (';'.join(','.join(_.name for _ in ti) for ti in getattr(n, TIPS_INSIDE))),
                                              ' '.join('"{}"'.format(_) for _ in get_states(n, columns))))


def save_to_pajek(nodes, edges, pajek):
    """
    Saves a compressed tree into Pajek format:

    *vertices <number_of_vertices>
    <id_1> "<vertex_name>" "<tips_inside>" "<column1>:<state(s)>" ["<column2>:<state(s)>" ...]
    ...
    *arcs
    <source_id> <target_id> <weight>
    ...

    <tips_inside> list tips that were vertically compressed inside this node: they are comma-separated.
    If the node was also horizontally merged (i.e. represents several similar configurations),
    the tip sets corresponding to different configurations are semicolon-separated.

    <state(s)> lists the states predicted for the corresponding column:
    if there are several states, they are separated with " or ".


    :return: void (creates a file specified in the pajek argument)
    """
    with open(pajek, 'w+') as f:
        f.write('*vertices {}\n'.format(len(nodes)))
        f.write('\n'.join(nodes))
        f.write('\n')
        f.write('*arcs\n')
        f.write('\n'.join(edges))


def compress_tree(tree, columns, can_merge_diff_sizes=True, tip_size_threshold=REASONABLE_NUMBER_OF_TIPS, mixed=False,
                  pajek=None, pajek_timing=VERTICAL):
    compressed_tree = copy_forest([tree], features=columns | set(tree.features))[0]

    for n_compressed, n in zip(compressed_tree.traverse('postorder'), tree.traverse('postorder')):
        n_compressed.add_feature(TIPS_BELOW, [list(n_compressed.iter_leaves())])
        n_compressed.add_feature(TIPS_INSIDE, [])
        n_compressed.add_feature(INTERNAL_NODES_INSIDE, [])
        n_compressed.add_feature(ROOTS, [n])
        if n_compressed.is_leaf():
            getattr(n_compressed, TIPS_INSIDE).append(n)
        elif not getattr(n_compressed, IS_POLYTOMY, False):
            getattr(n_compressed, INTERNAL_NODES_INSIDE).append(n)
        n.add_feature(COMPRESSED_NODE, n_compressed)

    collapse_vertically(compressed_tree, columns, mixed=mixed)
    if pajek is not None and VERTICAL == pajek_timing:
        _tree2pajek_vertices_arcs(compressed_tree, *pajek, columns=sorted(columns))

    for n in compressed_tree.traverse():
        n.add_feature(NUM_TIPS_INSIDE, len(getattr(n, TIPS_INSIDE)))
        n.add_feature(TIPS_INSIDE, [getattr(n, TIPS_INSIDE)])
        n.add_feature(INTERNAL_NODES_INSIDE, [getattr(n, INTERNAL_NODES_INSIDE)])

    get_bin = lambda _: _
    collapse_horizontally(compressed_tree, columns, get_bin, mixed=mixed)

    if can_merge_diff_sizes and len(compressed_tree) > tip_size_threshold:
        get_bin = lambda _: int(np.log10(max(1, _)))
        logging.getLogger('pastml').debug('Allowed merging nodes of different sizes.')
        collapse_horizontally(compressed_tree, columns, get_bin, mixed=mixed)

    if pajek is not None and HORIZONTAL == pajek_timing:
        _tree2pajek_vertices_arcs(compressed_tree, *pajek, columns=sorted(columns))

    if len(compressed_tree) > tip_size_threshold:
        for n in compressed_tree.traverse('preorder'):
            multiplier = (getattr(n.up, 'multiplier') if n.up else 1) * len(getattr(n, ROOTS))
            n.add_feature('multiplier', multiplier)

        def get_tsize(n):
            if getattr(n, IN_FOCUS, False) or getattr(n, AROUND_FOCUS, False) or getattr(n, UP_FOCUS, False):
                return np.inf
            return getattr(n, NUM_TIPS_INSIDE) * getattr(n, 'multiplier')

        node_thresholds = []
        for n in compressed_tree.traverse('postorder'):
            children_bs = 0 if not n.children else max(get_tsize(_) for _ in n.children)
            bs = get_tsize(n)
            # if bs > children_bs it means that the trimming threshold for the node is higher
            # than the ones for its children
            if not n.is_root() and bs > children_bs:
                node_thresholds.append(bs)
        threshold = sorted(node_thresholds)[-tip_size_threshold]

        if min(node_thresholds) >= threshold:
            if threshold == np.inf:
                logging.getLogger('pastml') .debug('All tips are in focus.')
            else:
                logging.getLogger('pastml')\
                    .debug('No tip is smaller than the threshold ({}, the size of the {}-th largest tip).'
                           .format(threshold, tip_size_threshold))
        else:
            if threshold == np.inf:
                logging.getLogger('pastml')\
                    .debug('Removing all the out of focus tips (as there are at least {} tips in focus).'
                           .format(tip_size_threshold))
            else:
                logging.getLogger('pastml').debug('Set tip size threshold to {} (the size of the {}-th largest tip).'
                                                  .format(threshold, tip_size_threshold))
            remove_small_tips(compressed_tree=compressed_tree, full_tree=tree,
                              to_be_removed=lambda _: get_tsize(_) < threshold)
            remove_mediators(compressed_tree, columns)
            collapse_horizontally(compressed_tree, columns, get_bin, mixed=mixed)

    if pajek is not None and TRIM == pajek_timing:
        _tree2pajek_vertices_arcs(compressed_tree, *pajek, columns=sorted(columns))

    return compressed_tree


def collapse_horizontally(tree, columns, tips2bin, mixed=False):
    config_cache = {}

    def get_configuration(n):
        if n.name not in config_cache:
            # Configuration is (branch_width, (size, states, child_configurations)),
            # where branch_width is only used for recursive calls and is ignored when considering a merge
            config_cache[n.name] = (len(getattr(n, TIPS_INSIDE)),
                               (tips2bin(getattr(n, NUM_TIPS_INSIDE)),
                                tuple(tuple(sorted(getattr(n, column, set()))) for column in columns),
                                tuple(sorted([get_configuration(_) for _ in n.children]))))
        return config_cache[n.name]

    collapsed_configurations = 0

    uncompressable_ids = set()
    for n in tree.traverse('postorder'):
        config2children = defaultdict(list)
        for _ in n.children:
            if mixed and (getattr(_, IN_FOCUS, False) or _.name in uncompressable_ids):
                uncompressable_ids.add(_.name)
                uncompressable_ids.add(n.name)
            else:
                # use (size, states, child_configurations) as configuration (ignore branch width)
                config2children[get_configuration(_)[1]].append(_)
        for children in (_ for _ in config2children.values() if len(_) > 1):
            collapsed_configurations += 1
            child = children[0]
            for sibling in children[1:]:
                getattr(child, TIPS_INSIDE).extend(getattr(sibling, TIPS_INSIDE))
                for ti in getattr(sibling, TIPS_INSIDE):
                    for _ in ti:
                        _.add_feature(COMPRESSED_NODE, child)
                getattr(child, INTERNAL_NODES_INSIDE).extend(getattr(sibling, INTERNAL_NODES_INSIDE))
                for ii in getattr(sibling, INTERNAL_NODES_INSIDE):
                    for _ in ii:
                        _.add_feature(COMPRESSED_NODE, child)
                getattr(child, ROOTS).extend(getattr(sibling, ROOTS))
                getattr(child, TIPS_BELOW).extend(getattr(sibling, TIPS_BELOW))
                n.remove_child(sibling)
            child.add_feature(METACHILD, True)
            child.add_feature(NUM_TIPS_INSIDE,
                              sum(len(_) for _ in getattr(child, TIPS_INSIDE)) / len(getattr(child, TIPS_INSIDE)))
            if child.name in config_cache:
                config_cache[child.name] = (len(getattr(child, TIPS_INSIDE)), config_cache[child.name][1])
    if collapsed_configurations:
        logging.getLogger('pastml').debug(
            'Collapsed {} sets of equivalent configurations horizontally.'.format(collapsed_configurations))


def remove_small_tips(compressed_tree, full_tree, to_be_removed):
    num_removed = 0
    changed = True
    while changed:
        changed = False
        for l in compressed_tree.get_leaves():
            parent = l.up
            if parent and to_be_removed(l):
                num_removed += 1
                parent.remove_child(l)
                # remove the corresponding nodes from the non-collapsed tree
                for ti in getattr(l, TIPS_INSIDE):
                    for _ in ti:
                        _.up.remove_child(_)
                for ii in getattr(l, INTERNAL_NODES_INSIDE):
                    for _ in ii:
                        _.up.remove_child(_)
                changed = True

    # if the full tree now contains non-sampled tips,
    # remove them from the tree and from the corresponding collapsed nodes
    todo = list(full_tree)
    while todo:
        t = todo.pop()
        if not getattr(t, IS_TIP, False):
            parent = t.up
            t.up.remove_child(t)
            if parent.is_leaf():
                todo.append(parent)
            for ini_list in getattr(getattr(t, COMPRESSED_NODE), INTERNAL_NODES_INSIDE):
                if t in ini_list:
                    ini_list.remove(t)

    logging.getLogger('pastml').debug(
        'Recursively removed {} tips of size smaller than the threshold.'.format(num_removed))


def collapse_vertically(tree, columns, mixed=False):
    """
    Collapses a child node into its parent if they are in the same state.
    :param columns: a list of characters
    :param tree: ete3.Tree
    :param mixed: if True then the nodes in focus will not get collapsed
    :return: void, modifies the input tree
    """

    def _same_states(node1, node2, columns):
        for column in columns:
            if getattr(node1, column, set()) != getattr(node2, column, set()):
                return False
        if mixed:
            if getattr(node1, IN_FOCUS, False) or getattr(node2, IN_FOCUS, False):
                return False
            if getattr(node1, UP_FOCUS, False) and not getattr(node2, IN_FOCUS, False) and not getattr(node2, UP_FOCUS, False):
                node2.add_feature(AROUND_FOCUS, True)
                return False
            if getattr(node2, UP_FOCUS, False) and not getattr(node1, IN_FOCUS, False) and not getattr(node1, UP_FOCUS, False):
                node1.add_feature(AROUND_FOCUS, True)
                return False
        return True

    num_collapsed = 0
    for n in tree.traverse('postorder'):
        if n.is_leaf():
            continue

        children = list(n.children)
        for child in children:
            # merge the child into this node if their states are the same
            if _same_states(n, child, columns):
                getattr(n, TIPS_INSIDE).extend(getattr(child, TIPS_INSIDE))
                for _ in getattr(child, TIPS_INSIDE):
                    _.add_feature(COMPRESSED_NODE, n)
                getattr(n, INTERNAL_NODES_INSIDE).extend(getattr(child, INTERNAL_NODES_INSIDE))
                for _ in getattr(child, INTERNAL_NODES_INSIDE):
                    _.add_feature(COMPRESSED_NODE, n)

                n.remove_child(child)
                grandchildren = list(child.children)
                for grandchild in grandchildren:
                    n.add_child(grandchild)
                num_collapsed += 1
    if num_collapsed:
        logging.getLogger('pastml').debug('Collapsed vertically {} internal nodes without state change.'
                                          .format(num_collapsed))


def remove_mediators(tree, columns):
    """
    Removes intermediate nodes that are just mediators between their parent and child states.
    :param columns: list of characters
    :param tree: ete3.Tree
    :return: void, modifies the input tree
    """
    num_removed = 0
    for n in tree.traverse('postorder'):
        if getattr(n, METACHILD, False) or n.is_leaf() or len(n.children) > 1 or n.is_root() \
                or getattr(n, NUM_TIPS_INSIDE) > 0:
            continue

        parent = n.up
        child = n.children[0]

        compatible = True
        for column in columns:
            states = getattr(n, column, set())
            parent_states = getattr(parent, column, set())
            child_states = getattr(child, column, set())
            # if mediator has unresolved states, it should hesitate between the parent and the child:
            if len(states) < 2 or states != child_states | parent_states:
                compatible = False
                break

        if compatible:
            parent.remove_child(n)
            parent.add_child(child)
            # update the uncompressed tree
            for ii in getattr(n, INTERNAL_NODES_INSIDE):
                for _ in ii:
                    for c in list(_.children):
                        _.up.add_child(c)
                    _.up.remove_child(_)
            num_removed += 1
    if num_removed:
        logging.getLogger('pastml').debug("Removed {} internal node{}"
                                          " with the state unresolved between the parent's and the only child's."
                                          .format(num_removed, '' if num_removed == 1 else 's'))
